import typing as T
from abc import ABC

import torch

from .base.models import LearningStep, PpoHyperParams, TrainingStep
from .monte_carlo_advantage_actor_critic import MonteCarloA2c
from .replay_buffers import ReplayBufferEntry


class PpoAgent(MonteCarloA2c, ABC):
    def __init__(self,
                 action_space: int,
                 agent_params: PpoHyperParams = PpoHyperParams(),
                 use_gpu: bool = True):
        super(MonteCarloA2c, self).__init__(action_space, agent_params, use_gpu)
        self.hyper_params = agent_params
        self.actor_critic_new = self.build_model().to(self.device).eval()
        self.actor_critic_new_optimizer = torch.optim.Adam(self.actor_critic_new.parameters(), lr=agent_params.lr)
        self.actor_critic_new.load_state_dict(self.actor_critic.state_dict())
        self.add_step_callback("ppo agent update state", self.ensure_learning_step_callback)

    def get_state_dict(self) -> dict:
        state_dict = super(PpoAgent, self).get_state_dict()
        state_dict.update({"actor_critic_new": self.actor_critic_new.state_dict()})
        return state_dict

    def ensure_learning_step_callback(self, training_step: TrainingStep) -> None:
        if training_step.step % self.hyper_params.ensure_every == 0:
            self.actor_critic.load_state_dict(self.actor_critic_new.state_dict())

    def learn(self, batch: T.List[ReplayBufferEntry]) -> None:
        batch_s = torch.stack([self.preprocess(m.s) for m in batch], 0).to(self.device).requires_grad_(True)
        batch_a = torch.tensor([m.a for m in batch], device=self.device)
        batch_rt = torch.tensor([[m.r] for m in batch], dtype=torch.float32, device=self.device)
        batch_weights = torch.tensor([m.weight for m in batch], device=self.device)

        new_action_probabilities, state_values = self.actor_critic_new(batch_s)

        new_chosen_action_log_probabilities_list = []
        new_chosen_action_log_entropy_list = []

        for p, a in zip(new_action_probabilities, batch_a):
            dist = torch.distributions.Categorical(p)
            new_chosen_action_log_probabilities_list.append(dist.log_prob(a))
            new_chosen_action_log_entropy_list.append(dist.entropy())

        new_chosen_action_log_probabilities: torch.Tensor = torch.stack(new_chosen_action_log_probabilities_list)
        new_chosen_action_log_entropy: torch.Tensor = torch.stack(new_chosen_action_log_entropy_list)

        with torch.no_grad():
            old_action_probabilities, _ = self.actor_critic(batch_s)
            old_chosen_action_log_probabilities: torch.Tensor = torch.stack(
                [torch.distributions.Categorical(p).log_prob(a) for p, a in zip(old_action_probabilities, batch_a)])

        advantages: torch.Tensor = (batch_rt - state_values.detach()).squeeze(1)

        ratios = torch.exp(new_chosen_action_log_probabilities-old_chosen_action_log_probabilities)
        surrogate_loss_1 = ratios * advantages
        surrogate_loss_2 = torch.clamp(ratios, 1 - self.hyper_params.clip_factor, 1 + self.hyper_params.clip_factor) * advantages

        actor_loss: torch.Tensor = -torch.min(surrogate_loss_1, surrogate_loss_2) * batch_weights
        critic_loss: torch.Tensor = self.loss_f(state_values, batch_rt) * batch_weights
        entropy_loss: torch.Tensor = -self.hyper_params.entropy_factor * new_chosen_action_log_entropy * batch_weights

        loss = (actor_loss + critic_loss + entropy_loss).mean()

        self.actor_critic_new_optimizer.zero_grad()
        loss.backward()
        self.actor_critic_new_optimizer.step()
        self.call_learn_callbacks(LearningStep(batch, [v.item() for v in state_values], [v.item() for v in batch_rt]))



import random
from .base.replay_buffer import ReplayBuffer
from .base.params import RandomReplayBufferParams
from abc import ABC


class RandomReplayBuffer(ReplayBuffer, ABC):
    def __init__(self, replay_buffer_params: RandomReplayBufferParams = RandomReplayBufferParams(), *args, **kwargs):
        if not isinstance(replay_buffer_params, RandomReplayBufferParams):
            raise ValueError("argument rp must be an instance or subclass of RandomReplayBufferParams")
        super(RandomReplayBuffer, self).__init__(replay_buffer_params, *args, **kwargs)

    def rp_sample(self, limit: int):
        indexes = random.sample(list(range(self.rp_get_length())), limit)
        return [self.records[i] for i in indexes]

    def rp_link(self):
        pass



import typing as T
from abc import ABC

import numpy as np
import torch
import torch.nn.functional as F

from ..environments import Environment
from .base.base_agent import BaseAgent
from .base.models import A2CHyperParams, TrainingProgress, TrainingParams, LearningStep, TrainingStep
from .explorers.noisy_explorer import NoisyLinear
from .replay_buffers import ReplayBufferEntry


class ActorCritic(torch.nn.Module):
    def __init__(self, model: torch.nn.Module, action_size: int, last_layer_factory: T.Callable[[int, int], torch.nn.Module]):
        super(ActorCritic, self).__init__()
        self.model = model
        last_layer = list(model.modules())[-1]
        if not isinstance(last_layer, (torch.nn.Linear, NoisyLinear)):
            raise ValueError("the model you have created must have a torch.nn.Linear or "
                             "agents.explorers.noisy_explorer.NoisyLinear in the last layer")

        if last_layer.out_features == action_size:
            print("WARNING: detected same number of features in the output of the model than the action space")

        self.actor = last_layer_factory(last_layer.out_features, action_size)
        self.critic = last_layer_factory(last_layer.out_features, 1)

    def forward(self, x, use_critic: bool = True):
        x = self.model(x)
        return F.softmax(self.actor(x), dim=1), (self.critic(x) if use_critic else None)


class A2cAgent(BaseAgent, ABC):
    def __init__(self,
                 action_space: int,
                 agent_params: A2CHyperParams = A2CHyperParams(),
                 use_gpu: bool = True):
        super(A2cAgent, self).__init__(action_space, agent_params, use_gpu)
        self.actor_critic = self.build_model().to(self.device)
        self.actor_critic_optimizer = torch.optim.Adam(self.actor_critic.parameters(), lr=agent_params.lr)
        self.loss_f = torch.nn.MSELoss(reduction="none").to(self.device)

    def get_info(self) -> dict:
        return {}

    def get_state_dict(self) -> dict:
        return {"actor_critic": self.actor_critic.state_dict()}

    def agent_specification_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        self.log.info("wrapping model with actor critic layer")
        return ActorCritic(model, self.action_space, self.last_layer_factory)

    def postprocess(self, t: torch.Tensor) -> np.ndarray:
        return np.array(t.squeeze(0))

    def infer(self, x: np.ndarray) -> np.ndarray:
        with torch.no_grad():
            preprocessed = self.preprocess(x).unsqueeze(0).to(self.device)
            infer = self.actor_critic.forward(preprocessed, use_critic=False)
            return self.postprocess(infer[0].cpu())

    def learn(self, batch: T.List[ReplayBufferEntry]) -> None:
        batch_s = torch.stack([self.preprocess(m.s) for m in batch], 0).to(self.device).requires_grad_(True)
        batch_s_ = torch.stack([self.preprocess(m.s_) for m in batch], 0).to(self.device)
        batch_a = torch.tensor([m.a for m in batch], device=self.device)
        batch_r = torch.tensor([[m.r] for m in batch], dtype=torch.float32, device=self.device)
        batch_finals = torch.tensor([[int(not m.final)] for m in batch], device=self.device)
        batch_weights = torch.tensor([m.weight for m in batch], device=self.device)

        action_probabilities, state_values = self.actor_critic(batch_s)
        with torch.no_grad():
            _, next_state_values = self.actor_critic(batch_s_)
            estimated_q_value: torch.Tensor = self.agent_params.gamma * next_state_values * batch_finals + batch_r

        advantages: torch.Tensor = (estimated_q_value - state_values.clone().detach()).squeeze(1)
        chosen_action_log_probabilities: torch.Tensor = torch.stack(
            [torch.distributions.Categorical(p).log_prob(a) for p, a in zip(action_probabilities, batch_a)])
        actor_loss: torch.Tensor = -chosen_action_log_probabilities * advantages * batch_weights
        critic_loss: torch.Tensor = self.loss_f(state_values, estimated_q_value) * batch_weights
        loss = (actor_loss + critic_loss).mean()
        self.actor_critic_optimizer.zero_grad()
        loss.backward()
        self.actor_critic_optimizer.step()
        self.call_learn_callbacks(LearningStep(batch, [v.item() for v in state_values], [v.item() for v in estimated_q_value]))

    def train(self, env: Environment, tp: TrainingParams = None) -> None:
        if tp is None:
            tp = self.default_training_params
        self.health_check(env)
        s = env.reset()
        i = 0
        episode = 1
        steps_survived = 0
        accumulated_reward = 0
        while True:
            estimated_rewards = self.infer(s)
            def choosing_f(x): return torch.distributions.Categorical(torch.tensor(x)).sample().item()
            a = self.ex_choose(list(estimated_rewards), choosing_f)
            s_, r, final = env.step(a)
            self.rp_add(ReplayBufferEntry(s, s_, a, r, final))
            accumulated_reward += r
            s = s_

            self.call_step_callbacks(TrainingStep(i, episode))

            if i % self.agent_params.learn_every == 0 and i != 0 and self.rp_get_length() >= tp.batch_size:
                batch = self.rp_sample(tp.batch_size)
                self.learn(batch)

            if final:
                training_progress = TrainingProgress(i, episode, steps_survived, accumulated_reward)
                must_exit = self.call_progress_callbacks(training_progress)
                if episode >= tp.episodes or must_exit:
                    return

                accumulated_reward = 0
                steps_survived = 0
                episode += 1
                s = env.reset()

            else:
                steps_survived += 1
            env.render()
            i += 1



from abc import ABC, abstractmethod
import typing as T

T_entry = T.TypeVar("T_entry")


class ReplayBufferInterface(T.Generic[T_entry], ABC):
    @abstractmethod
    def rp_get_length(self):
        raise NotImplementedError()

    @abstractmethod
    def rp_clear(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def rp_add(self, entry: T_entry) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def rp_sample(self, limit: int) -> T.List[T_entry]:
        raise NotImplementedError()

    @abstractmethod
    def rp_link(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def rp_get_stats(self) -> T.Dict[str, float]:
        raise NotImplementedError()



from abc import ABC, abstractmethod
import typing as T


T_S = T.TypeVar("T_S")


class Environment(T.Generic[T_S], ABC):
    @abstractmethod
    def reset(self) -> T_S:
        raise NotImplementedError()

    @abstractmethod
    def step(self, action: int) -> T.Tuple[T_S, float, bool]:
        raise NotImplementedError()

    @abstractmethod
    def render(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError()



from abc import ABC
import typing as T
from .params import ReplayBufferEntry, ReplayBufferParams
from ...base.base_agent import BaseAgent


class ReplayBuffer(BaseAgent, ABC):
    def __init__(self, replay_buffer_params: ReplayBufferParams, *args, **kwargs):
        super(ReplayBuffer, self).__init__(*args, **kwargs)
        self.rp: ReplayBufferParams = replay_buffer_params
        self.records: T.List[T.Union[None, ReplayBufferEntry]] = [None for _ in range(self.rp.max_len)]
        self.ptr = 0
        self.filled = False

    def rp_get_length(self):
        return self.ptr if not self.filled else self.rp.max_len

    def rp_clear(self) -> None:
        self.log.info("buffer is being emptied")
        self.records = [None for _ in range(self.rp.max_len)]
        self.ptr = 0
        self.filled = False

    def rp_add(self, entry: T.Union[ReplayBufferEntry]) -> bool:
        self.log.debug("adding new entry to bufer")
        self.records[self.ptr] = entry
        entry.index = self.ptr
        self.ptr = (self.ptr + 1) % self.rp.max_len
        if not self.filled and self.ptr == 0:
            self.log.info("buffer has been filled")
            self.filled = True
        return True

    def rp_get_stats(self) -> T.Dict[str, float]:
        return {"Replay Buffer Size": self.rp_get_length()}



import typing as T
from .random_replay_buffer import RandomReplayBuffer
from .n_steps_random_replay_buffer import NStepsRandomReplayBuffer
from .priorized_replay_buffer import PrioritizedReplayBuffer
from .n_steps_priorized_replay_buffer import NStepsPrioritizedReplayBuffer
from .base.params import ReplayBufferParams, RandomReplayBufferParams, ReplayBufferEntry, MemoryReplayBufferEntry, \
    NStepReplayBufferParams, NStepPrioritizedReplayBufferParams, PrioritizedReplayBufferParams,\
    NStepRandomReplayBufferParams

AnyReplayBuffer = T.Union[RandomReplayBuffer,
                          NStepsRandomReplayBuffer,
                          NStepsPrioritizedReplayBuffer,
                          PrioritizedReplayBuffer]



import numpy as np

from . import tools
from ...replay_buffers import PrioritizedReplayBuffer, PrioritizedReplayBufferParams, ReplayBufferEntry


def is_prioritized(replay_buffer: PrioritizedReplayBuffer):
    important_index = 2
    for i in range(10):
        replay_buffer.rp_add(ReplayBufferEntry(np.array([i]), np.array([i]), i, i, False))

    replay_buffer._refactor_priorities(list(range(10)), [0.5 if i == important_index else 0.1 for i in range(10)])

    dist = {}
    for _ in range(100):
        samples = replay_buffer.rp_sample(5)
        assert len(samples) == 5
        for sample in samples:
            dist[sample.a] = 1 if sample.a not in dist else dist[sample.a] + 1

    for i in dist:
        if i == important_index:
            continue
        assert dist[important_index] > dist[i]

    last_index = 11
    replay_buffer._refactor_priorities(list(range(10)), [0.1 for _ in range(10)])
    replay_buffer.rp_add(ReplayBufferEntry(np.array([last_index]), np.array([last_index]), last_index, last_index, False))
    dist = {}
    for _ in range(100):
        samples = replay_buffer.rp_sample(3)
        assert len(samples) == 3
        for sample in samples:
            dist[sample.a] = 1 if sample.a not in dist else dist[sample.a] + 1

    for i in dist:
        if i == last_index:
            continue
        assert dist[last_index] > dist[i]


class Agent(PrioritizedReplayBuffer, tools.Agent):
    pass


def test_important_transitions_are_prioritized():
    prb = PrioritizedReplayBufferParams(max_len=15, alpha=0.5, init_beta=0.9, final_beta=1.0, increase_beta=1+1e-5)
    test_agent = Agent(replay_buffer_params=prb)
    is_prioritized(test_agent)



import typing as T
from abc import ABC
from collections import deque

from .replay_buffer import ReplayBuffer
from .params import ReplayBufferEntry, NStepReplayBufferParams


class NStepsReplayBuffer(ReplayBuffer, ABC):
    def __init__(self, replay_buffer_params: NStepReplayBufferParams = NStepReplayBufferParams(), *args, **kwargs):
        self.rp: NStepReplayBufferParams = replay_buffer_params
        self.n_step_buffer: T.Deque[ReplayBufferEntry] = deque(maxlen=self.rp.n_step)
        super().__init__(replay_buffer_params, *args, **kwargs)

    def _get_n_step_info(self) -> ReplayBufferEntry:
        first_entry = self.n_step_buffer[0]
        last_entry = self.n_step_buffer[-1]
        ac_r, ac_s_, ac_final = last_entry.r, last_entry.s_, last_entry.final

        for transition in reversed(list(self.n_step_buffer)[:-1]):
            r, s_, final = transition.r, transition.s_, transition.final
            if self.accumulate_rewards:
                ac_r = r + self.agent_params.gamma * ac_r

            if final:
                ac_s_, ac_final, ac_r = (s_, final, r)

        return ReplayBufferEntry(first_entry.s, ac_s_, first_entry.a, ac_r, ac_final)

    def rp_add(self, entry: T.Union[ReplayBufferEntry]) -> bool:
        self.n_step_buffer.append(entry)
        if len(self.n_step_buffer) < self.rp.n_step:
            return False

        n_step_entry = self._get_n_step_info()
        super(NStepsReplayBuffer, self).rp_add(n_step_entry)
        return True

    def rp_clear(self):
        super(NStepsReplayBuffer, self).rp_clear()
        self.n_step_buffer.clear()

    def rp_link(self):
        prev_gamma = self.agent_params.gamma
        self.agent_params.gamma = self.agent_params.gamma ** self.rp.n_step
        self.log.info(f"refactoring gamma: {prev_gamma} -> {self.agent_params.gamma}")



import typing as T
from dataclasses import dataclass

import numpy as np


@dataclass
class ReplayBufferEntry:
    s: np.ndarray
    s_: np.ndarray
    a: int
    r: float
    final: bool
    index: T.Union[None, int] = None
    weight: int = 1


@dataclass
class AgentParams:
    gamma: float = 0.99
    learn_every: int = 1


@dataclass
class DqnHyperParams(AgentParams):
    lr: float = 0.0025


@dataclass
class DuelingDqnHyperParams(DqnHyperParams):
    pass


@dataclass
class DoubleDqnHyperParams(DqnHyperParams):
    ensure_every: int = 10


@dataclass
class DoubleDuelingDqnHyperParams(DoubleDqnHyperParams):
    pass


@dataclass
class A2CHyperParams(AgentParams):
    lr: float = 0.001


@dataclass
class PpoHyperParams(A2CHyperParams):
    clip_factor: float = 0.02
    entropy_factor: float = 0.01
    ensure_every: int = 10


@dataclass
class TrainingProgress:
    step: int
    episode: int
    steps_survived: int
    total_reward: float


TProgressCallback = T.Callable[[TrainingProgress], bool]


@dataclass
class TrainingParams:
    batch_size: int = 46
    episodes: int = 10000


@dataclass
class TrainingStep:
    step: int
    episode: int


TStepCallback = T.Callable[[TrainingStep], None]


@dataclass
class LearningStep:
    batch: T.List[T.Any]
    x: T.List[float]
    y: T.List[float]


TLearnCallback = T.Callable[[LearningStep], None]



import typing as T

import pytest
import torch
import torch.nn.functional as F

from ... import agents, environments

EXPECTED_REWARD = 25


class NN(torch.nn.Module):
    def __init__(self):
        super(NN, self).__init__()
        self.linear_1 = torch.nn.Linear(environments.CartPole.OBSERVATION_SPACE, 100)
        self.linear_2 = torch.nn.Linear(self.linear_1.out_features, 100)

    def forward(self, x):
        return F.relu(self.linear_2(F.relu(self.linear_1(x))))


params = []
for RB in [agents.replay_buffers.RandomReplayBuffer,
           agents.replay_buffers.NStepsPrioritizedReplayBuffer,
           agents.replay_buffers.PrioritizedReplayBuffer,
           agents.replay_buffers.NStepsPrioritizedReplayBuffer]:
    for EX in [agents.explorers.RandomExplorer,
               agents.explorers.NoisyExplorer]:
        for AG in [agents.DqnAgent,
                   agents.DoubleDqnAgent,
                   agents.DuelingDqnAgent,
                   agents.DoubleDuelingDqnAgent,
                   agents.A2cAgent,
                   agents.MonteCarloA2c,
                   agents.PpoAgent]:

            class Agent(RB, EX, AG):
                def model_factory(self) -> torch.nn.Module:
                    return NN()

                def preprocess(self, x):
                    return torch.from_numpy(x.astype("float32"))

            ag: T.Type[agents.AnyAgent] = Agent
            params.append(pytest.param(ag, id=f"{AG.__name__}({EX.__name__}, {RB.__name__})"))


@pytest.mark.parametrize("agent_class", params)
def test_agents(agent_class: T.Type[agents.AnyAgent]):
    env = environments.CartPole()
    env.visualize = False
    agent = agent_class(action_space=environments.CartPole.ACTION_SPACE)

    record: T.List[agents.TrainingProgress] = []

    def on_progress(progress: agents.TrainingProgress):
        record.append(progress)
        print(progress.total_reward)
        return progress.total_reward > EXPECTED_REWARD

    agent.add_progress_callback("testing", on_progress)
    agent.train(env, agents.TrainingParams(batch_size=64, episodes=200))
    assert any([r.total_reward > EXPECTED_REWARD for r in record])



from dataclasses import dataclass
import typing as T
import numpy as np
import torch


@dataclass
class ReplayBufferEntry:
    s: np.ndarray
    s_: np.ndarray
    a: int
    r: float
    final: bool
    index: T.Union[None, int] = None
    weight: int = 1


@dataclass
class MemoryReplayBufferEntry(ReplayBufferEntry):
    m: torch.Tensor = None


@dataclass
class ReplayBufferParams:
    max_len: int = 10000


@dataclass
class RandomReplayBufferParams(ReplayBufferParams):
    pass


@dataclass
class NStepReplayBufferParams(ReplayBufferParams):
    n_step: int = 3


@dataclass
class PrioritizedReplayBufferParams(ReplayBufferParams):
    alpha: float = 0.6
    init_beta: float = 0.4
    final_beta: float = 1.0
    increase_beta: float = 1e-4


@dataclass
class NStepPrioritizedReplayBufferParams(NStepReplayBufferParams, PrioritizedReplayBufferParams):
    pass


@dataclass
class NStepRandomReplayBufferParams(NStepReplayBufferParams, RandomReplayBufferParams):
    pass



from setuptools import setup, find_packages

setup(
    name='lazaro',
    version=open("version.txt").read(),
    packages=find_packages(),
    url='https://github.com/GabrielMusat/lazaro',
    python_requires=">=3.7.*",
    license='Apache 2.0',
    author='Gabriel Musat',
    install_requires=[req for req in open("requirements.txt").read().split("\n") if len(req) > 0],
    author_email='gabimtme@gmail.com',
    description='Reinforcement learning framework for implementing custom models on custom environments using state '
                'of the art RL algorithms',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: GPU",
        "Environment :: GPU :: NVIDIA CUDA",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Typing :: Typed"
    ]
)



from .cart_pole import CartPole
from .beam_rider import BeamRider
from .cart_pole_frames import CartPoleFrames
from .space_invaders import SpaceInvaders
from .space_invaders_ram import SpaceInvadersRam
from .frozenlake import FrozenLake
from .environment import Environment



import typing as T
from abc import ABC

import numpy as np
import torch

from ..environments import Environment
from .base.base_agent import BaseAgent
from .base.models import DqnHyperParams, TrainingParams, TrainingProgress, LearningStep, TrainingStep
from .explorers.noisy_explorer import NoisyLinear
from .replay_buffers import ReplayBufferEntry


class DqnNetwork(torch.nn.Module):
    def __init__(self,
                 model: torch.nn.Module,
                 action_space: int,
                 last_layer_factory: T.Callable[[int, int], torch.nn.Module]):
        super(DqnNetwork, self).__init__()
        self.model = model
        last_layer = list(model.modules())[-1]
        if not isinstance(last_layer, (torch.nn.Linear, NoisyLinear)):
            raise ValueError("the model you have created must have a torch.nn.Linear or "
                             "agents.explorers.noisy_explorer.NoisyLinear in the last layer")

        if last_layer.out_features == action_space:
            print("WARNING: detected same number of features in the output of the model than the action space")

        self.head = last_layer_factory(last_layer.out_features, action_space)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)
        return self.head(x)


class DqnAgent(BaseAgent, ABC):
    def __init__(self,
                 action_space: int,
                 agent_params: DqnHyperParams = DqnHyperParams(),
                 use_gpu: bool = True):
        super(DqnAgent, self).__init__(action_space, agent_params, use_gpu)
        self.action_estimator = self.build_model().to(self.device)
        self.optimizer = torch.optim.Adam(self.action_estimator.parameters(), lr=agent_params.lr)
        self.loss_f = torch.nn.MSELoss(reduction="none").to(self.device)

    def get_state_dict(self) -> dict:
        return {"action_estimator": self.action_estimator.state_dict()}

    def get_info(self) -> dict:
        return {}

    def agent_specification_model_modifier(self, model: torch.nn.Module) -> torch.nn.Module:
        self.log.info("wrapping model with simple dqn layer")
        return DqnNetwork(model, self.action_space, self.last_layer_factory)

    def infer(self, x: np.ndarray) -> np.ndarray:
        with torch.no_grad():
            preprocessed = self.preprocess(x).unsqueeze(0).to(self.device)
            infer = self.action_estimator.forward(preprocessed)
            return self.postprocess(infer.cpu())

    def postprocess(self, t: torch.Tensor) -> np.ndarray:
        return np.array(t.squeeze(0))

    def learn(self, batch: T.List[ReplayBufferEntry]) -> None:
        batch_s = torch.stack([self.preprocess(m.s) for m in batch], 0).to(self.device).requires_grad_(True)
        batch_s_ = torch.stack([self.preprocess(m.s_) for m in batch], 0).to(self.device)
        batch_a = [m.a for m in batch]
        batch_r = torch.tensor([m.r for m in batch], dtype=torch.float32, device=self.device)
        batch_finals = torch.tensor([int(not m.final) for m in batch], device=self.device)
        batch_weights = torch.tensor([m.weight for m in batch], device=self.device)
        actions_estimated_values: torch.Tensor = self.action_estimator(batch_s)
        with torch.no_grad():
            actions_expected_values: torch.Tensor = self.action_estimator(batch_s_)

        x = torch.stack([t_s[t_a] for t_s, t_a in zip(actions_estimated_values, batch_a)])
        y = torch.max(actions_expected_values, 1)[0] * self.agent_params.gamma * batch_finals + batch_r
        element_wise_loss = self.loss_f(x, y)
        loss = (element_wise_loss * batch_weights).mean()
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        self.call_learn_callbacks(LearningStep(batch, [v.item() for v in x], [v.item() for v in y]))

    def train(self, env: Environment, tp: TrainingParams = None) -> None:
        if tp is None:
            tp = self.default_training_params
        self.health_check(env)
        s = env.reset()
        i = 0
        episode = 1
        steps_survived = 0
        accumulated_reward = 0
        while True:
            estimated_rewards = self.infer(s)
            a = self.ex_choose(list(estimated_rewards), lambda x: np.argmax(estimated_rewards).item())
            s_, r, final = env.step(a)
            self.rp_add(ReplayBufferEntry(s, s_, a, r, final))
            accumulated_reward += r
            s = s_

            self.call_step_callbacks(TrainingStep(i, episode))

            if i % self.agent_params.learn_every == 0 and i != 0 and self.rp_get_length() >= tp.batch_size:
                batch = self.rp_sample(tp.batch_size)
                self.learn(batch)

            if final:
                training_progress = TrainingProgress(i, episode, steps_survived, accumulated_reward)
                must_exit = self.call_progress_callbacks(training_progress)
                if episode >= tp.episodes or must_exit:
                    return

                episode += 1
                accumulated_reward = 0
                steps_survived = 0
                s = env.reset()

            else:
                steps_survived += 1
            env.render()
            i += 1



import typing as T
from abc import ABC, abstractmethod


class ExplorerInterface(ABC):
    @abstractmethod
    def ex_choose(self, actions: T.List[float], f: T.Callable[[T.List[float]], int]) -> int:
        raise NotImplementedError()

    @abstractmethod
    def ex_link(self):
        raise NotImplementedError()

    @abstractmethod
    def ex_get_stats(self) -> T.Dict[str, float]:
        raise NotImplementedError()



import typing as T

import gym
import numpy as np

from .environment import Environment


class SpaceInvadersRam(Environment):
    def __init__(self):
        self.env: gym.Env = gym.make("SpaceInvaders-ram-v0")

    def reset(self) -> np.ndarray:
        return self.env.reset()

    def do_step(self, action: int) -> T.Tuple[np.ndarray, float, bool]:
        s, r, f, _ = self.env.step(action)
        return s, r, f

    def render(self) -> None:
        self.env.render()

    def close(self) -> None:
        self.env.close()
